from .BatsException import BatsException
from .BatsWarning import BatsWarning
from .InputArgsException import InputArgsException
from .InputArgsWarning import InputArgsWarning
from .ModelWarning import ModelWarning
from .ExceptionHandler import ExceptionHandler

__all__ = ['BatsException', 'BatsWarning', 'InputArgsException', 'InputArgsWarning', 'ModelWarning', 'ExceptionHandler']



import numpy as np
import multiprocessing as actual_processing
import multiprocessing.dummy as dummy_processing
from sklearn.base import BaseEstimator
from sklearn.utils.validation import check_array, column_or_1d as c1d
from sklearn.model_selection import ParameterGrid

import tbats.error as error


class Estimator(BaseEstimator):
    """Base estimator for BATS and TBATS models

    Methods
    -------
    fit(y)
        Fit to y and select best performing model based on AIC criterion.
    """

    def __init__(self, context,
                 use_box_cox=None, box_cox_bounds=(0, 1),
                 use_trend=None, use_damped_trend=None,
                 seasonal_periods=None, use_arma_errors=True,
                 n_jobs=None):
        """ Class constructor

        Parameters
        ----------
        context: abstract.ContextInterface
            For advanced users only. Provide this to override default behaviors
        use_box_cox: bool or None, optional (default=None)
            If Box-Cox transformation of original series should be applied.
            When None both cases shall be considered and better is selected by AIC.
        box_cox_bounds: tuple, shape=(2,), optional (default=(0, 1))
            Minimal and maximal Box-Cox parameter values.
        use_trend: bool or None, optional (default=None)
            Indicates whether to include a trend or not.
            When None both cases shall be considered and better is selected by AIC.
        use_damped_trend: bool or None, optional (default=None)
            Indicates whether to include a damping parameter in the trend or not.
            Applies only when trend is used.
            When None both cases shall be considered and better is selected by AIC.
        seasonal_periods: iterable or array-like, optional (default=None)
            Length of each of the periods (amount of observations in each period).
            BATS accepts only int values here.
            When None or empty array, non-seasonal model shall be fitted.
        use_arma_errors: bool, optional (default=True)
            When True BATS will try to improve the model by modelling residuals with ARMA.
            Best model will be selected by AIC.
            If False, ARMA residuals modeling will not be considered.
        show_warnings: bool, optional (default=True)
            If warnings should be shown or not.
            Also see Model.warnings variable that contains all model related warnings.
        n_jobs: int, optional (default=None)
            How many jobs to run in parallel when fitting BATS model.
            When not provided BATS shall try to utilize all available cpu cores.
        """
        self.context = context
        self.n_jobs = n_jobs

        self.seasonal_periods = self._normalize_seasonal_periods(seasonal_periods)
        self.use_box_cox = use_box_cox
        self.box_cox_bounds = box_cox_bounds
        self.use_arma_errors = use_arma_errors
        self.use_trend = use_trend
        if use_trend is False:
            if use_damped_trend is True:
                self.context.get_exception_handler().warn(
                    "When use_damped_trend can be used only with use_trend. Setting damped trend to False.",
                    error.InputArgsWarning
                )
            use_damped_trend = False
        self.use_damped_trend = use_damped_trend

    def _normalize_seasonal_periods(self, seasonal_periods):
        # abstract method
        raise NotImplementedError()

    def _do_fit(self, y):
        # abstract method
        raise NotImplementedError()

    def fit(self, y):
        """Fit model to observations ``y``.

        :param y: array-like or iterable, shape=(n_samples,)
        :return: abstract.Model, Fitted model
        """
        y = self._validate(y)
        if y is False:
            # Input data is not valid and no exception was raised yet.
            # This can happen only when one overrides default exception handler (see tbats.error.ExceptionHandler)
            return None

        if np.allclose(y, y[0]):
            return self.context.create_constant_model(y[0]).fit(y)

        best_model = self._do_fit(y)

        for warning in best_model.warnings:
            self.context.get_exception_handler().warn(warning, error.ModelWarning)

        return best_model

    def _validate(self, y):
        """Validates input time series. Also adjusts box_cox if necessary."""
        try:
            y = c1d(check_array(y, ensure_2d=False, force_all_finite=True, ensure_min_samples=1,
                                copy=True, dtype=np.float64))  # type: np.ndarray
        except Exception as validation_exception:
            self.context.get_exception_handler().exception(
                "y series is invalid", error.InputArgsException, previous_exception=validation_exception
            )
            return False

        if np.any(y <= 0):
            if self.use_box_cox is True:
                self.context.get_exception_handler().warn(
                    "Box-Cox transformation (use_box_cox) was forced to True "
                    "but there are negative values in input series. "
                    "Setting use_box_cox to False.",
                    error.InputArgsWarning
                )
            self.use_box_cox = False

        return y

    def _case_fit(self, components_combination):
        """Internal method used by parallel computation."""
        case = self.context.create_case_from_dictionary(**components_combination)
        return case.fit(self._y)

    def _choose_model_from_possible_component_settings(self, y, components_grid):
        """Fits all models in a grid and returns best one by AIC

        Returns
        -------
            abstract.Model
                Best model by AIC
        """
        self._y = y
        # note n_jobs = None means to use cpu_count()
        with self.context.multiprocessing().Pool(processes=self.n_jobs) as pool:
            models = pool.map(self._case_fit, components_grid)
            pool.close()
        self._y = None  # clean-up
        if len(models) == 0:
            return None
        best_model = models[0]
        for model in models:
            if model.aic < best_model.aic:
                best_model = model
        return best_model

    def _prepare_components_grid(self, seasonal_harmonics=None):
        """Provides a grid of all allowed model component combinations.

        Parameters
        ----------
        seasonal_harmonics: array-like or None
            When provided all component combinations shall contain those harmonics
        """
        allowed_combinations = []

        use_box_cox = self.use_box_cox

        base_combination = {
            'use_box_cox': self.__prepare_component_boolean_combinations(use_box_cox),
            'box_cox_bounds': [self.box_cox_bounds],
            'use_arma_errors': [self.use_arma_errors],
            'seasonal_periods': [self.seasonal_periods],
        }
        if seasonal_harmonics is not None:
            base_combination['seasonal_harmonics'] = [seasonal_harmonics]

        if self.use_trend is not True:  # False or None
            allowed_combinations.append({
                **base_combination,
                **{
                    'use_trend': [False],
                    'use_damped_trend': [False],  # Damped trend must be False when trend is False
                }
            })

        if self.use_trend is not False:  # True or None
            allowed_combinations.append({
                **base_combination,
                **{
                    'use_trend': [True],
                    'use_damped_trend': self.__prepare_component_boolean_combinations(self.use_damped_trend),
                }
            })
        return ParameterGrid(allowed_combinations)

    def _prepare_non_seasonal_components_grid(self):
        """Provides a grid of all allowed  non-season model component combinations."""
        allowed_combinations = []

        use_box_cox = self.use_box_cox

        base_combination = {
            'use_box_cox': self.__prepare_component_boolean_combinations(use_box_cox),
            'box_cox_bounds': [self.box_cox_bounds],
            'use_arma_errors': [self.use_arma_errors],
            'seasonal_periods': [[]],
        }

        if self.use_trend is not True:  # False or None
            allowed_combinations.append({
                **base_combination,
                **{
                    'use_trend': [False],
                    'use_damped_trend': [False],  # Damped trend must be False when trend is False
                }
            })

        if self.use_trend is not False:  # True or None
            allowed_combinations.append({
                **base_combination,
                **{
                    'use_trend': [True],
                    'use_damped_trend': self.__prepare_component_boolean_combinations(self.use_damped_trend),
                }
            })
        return ParameterGrid(allowed_combinations)

    @staticmethod
    def __prepare_component_boolean_combinations(param):
        combinations = [param]
        if param is None:
            combinations = [False, True]
        return combinations

    def _normalize_seasonal_periods_to_type(self, seasonal_periods, dtype):
        """Validates seasonal periods and normalizes them

        Normalization ensures periods are of proper type, unique and sorted.
        """
        if seasonal_periods is not None:
            try:
                seasonal_periods = c1d(check_array(seasonal_periods, ensure_2d=False, force_all_finite=True,
                                                   ensure_min_samples=0,
                                                   copy=True, dtype=dtype))
            except Exception as validation_exception:
                self.context.get_exception_handler().exception("seasonal_periods definition is invalid",
                                                               error.InputArgsException,
                                                               previous_exception=validation_exception)

            seasonal_periods = np.unique(seasonal_periods)
            if len(seasonal_periods[np.where(seasonal_periods <= 1)]) > 0:
                self.context.get_exception_handler().warn(
                    "All seasonal periods should be values greater than 1. "
                    "Ignoring all seasonal period values that do not meet this condition.",
                    error.InputArgsWarning
                )
            seasonal_periods = seasonal_periods[np.where(seasonal_periods > 1)]
            seasonal_periods.sort()
            if len(seasonal_periods) == 0:
                seasonal_periods = None
        return seasonal_periods



import numpy as np
from sklearn.utils.validation import check_array, column_or_1d as c1d

import tbats.error as error
from . import Guerrero


def find_box_cox_lambda(y, seasonal_periods=None, bounds=(-1, 2)):
    y = c1d(check_array(y, ensure_2d=False, force_all_finite=True, ensure_min_samples=1,
                        copy=False, dtype=np.float64))  # type: np.ndarray

    guerrero = Guerrero()
    return guerrero.find_lambda(y, seasonal_periods=seasonal_periods, bounds=bounds)


def boxcox(y, lam=None, seasonal_periods=None, bounds=(-1, 2)):
    y = c1d(check_array(y, ensure_2d=False, force_all_finite=True, ensure_min_samples=1,
                        copy=False, dtype=np.float64))  # type: np.ndarray
    if lam is None:
        lam = find_box_cox_lambda(y, seasonal_periods=seasonal_periods, bounds=bounds)
    if (lam <= 0 or np.isclose(lam, 0)) and np.any(y <= 0):
        raise error.InputArgsException('y must have only positive values for box-cox transformation.')
    if np.isclose(0.0, lam):
        return np.log(y)
    return (np.sign(y) * (np.abs(y) ** lam) - 1) / lam


def inv_boxcox(y, lam, force_valid=False):
    y = c1d(check_array(y, ensure_2d=False, force_all_finite=True, ensure_min_samples=1,
                        copy=False, dtype=np.float64))  # type: np.ndarray
    if np.isclose(0.0, lam):
        return np.exp(y)
    if lam < 0 and force_valid:
        y[y > -1 / lam] = -1 / lam
    if lam < 0 and np.any(y > -1 / lam):
        raise error.InputArgsException('Not possible to transform back such y values.')
    yy = y * lam + 1
    return np.sign(yy) * (np.abs(yy) ** (1 / lam))


import copy
import numpy as np

import tbats.transformation as transformation
from . import ArrayHelper


class ModelParams(object):
    """Holds all parameters needed to calculate model and forecasts

    Attributes
    ----------
    components: Components
        Defines which components are used in the model (trend, box-cox, seasonalities, ...)
    alpha: float
        Smoothing parameter value
    beta: float or None
        Trend smoothing parameter value.
        None when trend is not used.
        Also see components.use_trend.
    phi: float or None
        Trend damping parameter value.
        None when trend is not used. 1.0 when trend is used but damping is not.
        Also see components.use_damped_trend.
    box_cox_lambda: float or None
        Box-Cox transformation lambda parameter value
        None when series is not being transformed.
        Also see components.use_box_cox.
    gamma_params: array-like of floats
        Seasonal smoothing parameters.
        Empty array when there is no seasonality in the model.
        Also see components.seasonal_periods
    ar_coefs: array-like of floats
        AR(p) parameters used for residuals modeling.
        Empty array when there is no ARMA residuals modelling.
        Also see components.p and components.use_arma_errors.
    ma_coefs: array-like of floats or None
        MA(q) parameters used for residuals modeling.
        Empty array when there is no ARMA residuals modelling.
        Also see components.q and components.use_arma_errors.
    x0: array-like of floats or None
        Seed state for computations consisting of trend, seasonal and ARMA related seeds.
    """

    def __init__(self, components, alpha, beta=None, phi=None, box_cox_lambda=None,
                 gamma_params=None,
                 ar_coefs=None, ma_coefs=None, x0=None):
        """Holds all parameters needed to calculate model and forecasts

        Parameters
        ----------
        components: Components
            Defines which components are used in the model (trend, box-cox, seasonalities, ...)
        alpha: float
            Smoothing parameter value
        beta: float or None
            Trend smoothing parameter value.
            None when trend is not used.
            Also see components.use_trend.
        phi: float or None
            Trend damping parameter value.
            None when damping is not used.
            Also see components.use_damped_trend.
        box_cox_lambda: float or None
            Box-Cox transformation lambda parameter value
            None when series is not being transformed.
            Also see components.use_box_cox.
        gamma_params: array-like of floats or None
            Seasonal smoothing parameters.
            Also see components.seasonal_periods
        ar_coefs: array-like of floats or None
            AR(p) parameters used for residuals modeling.
            Also see components.p and components.use_arma_errors.
        ma_coefs: array-like of floats or None
            MA(q) parameters used for residuals modeling.
            Also see components.q and components.use_arma_errors.
        x0: array-like of floats or None
            Seed state for computations consisting of trend, seasonal and ARMA related seeds.
            When not provided will be initialized to a vector of zeroes.
        """
        self.components = components
        self.alpha = alpha
        self.box_cox_lambda = None
        if self.components.use_box_cox:
            self.box_cox_lambda = box_cox_lambda
        self.beta = None
        self.phi = None
        if self.components.use_trend:
            self.beta = beta
            self.phi = 1.
            if self.components.use_damped_trend:
                self.phi = phi
        self.gamma_params = self._normalize_gamma_params(gamma_params)

        self.__init_arma(ar_coefs, ma_coefs)

        self.__init_x0(x0)

    @classmethod
    def with_default_starting_params(cls, y, components):
        # abstract method
        raise NotImplementedError()

    def seasonal_components_amount(self):
        # abstract method
        raise NotImplementedError()

    def _normalize_gamma_params(self, gamma_params=None):
        # abstract method
        raise NotImplementedError()

    @classmethod
    def find_initial_box_cox_lambda(cls, y, components):
        """ Chooses starting Box-Cox lambda parameter using Guerrero method.

        Parameters
        ----------
        y: array-like
            Time series
        components: Components
            Components of the model

        Returns
        -------
        float or None
            Lambda parameter or None when no Box-Cox transformation applies.
        """
        if not components.use_box_cox:
            return None
        return transformation.find_box_cox_lambda(
            y,
            seasonal_periods=components.seasonal_periods,
            bounds=components.box_cox_bounds,
        )

    def with_arma(self, p=0, q=0):
        """Returns copy of itself with provided ARMA levels.

        ARMA coefficients are initialized to vectors of lengths p and q of zero values.

        Parameters
        ----------
        p: int
            Auto-regressive level p
        q: int
            Moving average level q

        Returns
        -------
        ModelParams
            Copy with ARMA of provided levels
        """
        me = copy.deepcopy(self)
        me.components = self.components.with_arma(p, q)
        me.__init_arma()
        return me

    def with_zero_x0(self):
        """Returns a copy of itself with seed state values set to zeroes.

        Returns
        -------
        ModelParams
            Copy with x0 as a vector of zeroes
        """
        me = copy.deepcopy(self)
        me.__init_x0()
        return me

    def with_x0(self, x0):
        """Returns a copy of itself with seed states set to provided values.

        Parameters
        ----------
        x0: array-like of floats
            Seed states

        Returns
        -------
        ModelParams
            Copy with x0 set to provided values
        """
        me = copy.deepcopy(self)
        me.__init_x0(x0)
        return me

    def with_vector_values(self, vector):
        """Returns a copy of itself with model parameters taken from the vector

        If necessary x0 is re-transformed to match new Box-Cox lambda parameter.

        Parameters
        ----------
        vector: array-like of floats
            A vector with model parameters. When all parameters are used it consist of:
            (alpha, box_cox_lambda, beta, phi, gamma parameters, AR coefficients, MA coefficients)
            When some components are not used they are also not present in the vector.

        Returns
        -------
        ModelParams
            A copy of itself with model parameters taken from the vector
        """
        x0 = self.x0.copy()

        alpha = vector[0]

        offset = 1
        boxcox = None
        if self.components.use_box_cox:
            boxcox = vector[offset]
            # recalculate x0 according to changed box-cox parameter
            x0 = transformation.boxcox(
                transformation.inv_boxcox(x0, self.box_cox_lambda),
                boxcox
            )

            offset += 1

        beta = None
        if self.components.use_trend:
            beta = vector[offset]
            offset += 1

        phi = None
        if self.components.use_damped_trend:  # note that damped trend is used only when trend is used
            phi = vector[offset]
            offset += 1

        gamma_params = vector[offset:(offset + self.components.gamma_params_amount())]  # may be empty
        offset += len(self.components.seasonal_periods)

        ar_coefs = vector[offset:(offset + self.components.p)]  # may be empty
        offset += self.components.p

        ma_coefs = vector[offset:(offset + self.components.q)]  # may be empty
        # offset += self.components.q

        return self.__class__(components=self.components,
                              alpha=alpha, beta=beta, phi=phi, box_cox_lambda=boxcox,
                              gamma_params=gamma_params,
                              ar_coefs=ar_coefs, ma_coefs=ma_coefs, x0=x0)

    def _create_x0_of_zeroes(self):
        """Creates seed vector of proper length of zeroes"""
        x0_length = 1
        if self.components.use_trend:
            x0_length += 1

        x0_length += self.seasonal_components_amount()

        x0_length += self.components.arma_length()

        return np.asarray([0] * x0_length)

    def to_vector(self):
        v = [[self.alpha]]
        if self.components.use_box_cox:
            v.append([self.box_cox_lambda])
        if self.components.use_trend:
            v.append([self.beta])
        if self.components.use_damped_trend:
            v.append([self.phi])
        if self.components.gamma_params_amount() > 0:
            v.append(self.gamma_params)
        if self.components.p > 0:
            v.append(self.ar_coefs)
        if self.components.q > 0:
            v.append(self.ma_coefs)
        return np.concatenate(v)

    def amount(self):
        """
        Amount of parameters in the model, including seed state
        :return: int
        """
        amount = 1
        if self.components.use_trend:
            amount += 1
        if self.components.use_damped_trend:
            amount += 1
        if self.components.use_box_cox:
            amount += 1

        return len(self.x0) + amount + self.components.gamma_params_amount() + self.components.arma_length()

    def is_box_cox_in_bounds(self):
        return self.components.is_box_cox_in_bounds(self.box_cox_lambda)

    def summary(self):
        s = self.components.summary()
        if not self.box_cox_lambda is None:
            s += 'Box-Cox Lambda %f\n' % self.box_cox_lambda
        s += 'Smoothing (Alpha): %f\n' % self.alpha
        if not self.beta is None:
            s += 'Trend (Beta): %f\n' % self.beta
        if not self.phi is None:
            s += 'Damping Parameter (Phi): %f\n' % self.phi
        s += 'Seasonal Parameters (Gamma): %s\n' % self.gamma_params
        s += 'AR coefficients %s\n' % self.ar_coefs
        s += 'MA coefficients %s\n' % self.ma_coefs
        s += 'Seed vector %s\n' % self.x0
        return s

    def __init_arma(self, ar_coefs=None, ma_coefs=None):
        self.ar_coefs = ArrayHelper.to_array(ar_coefs)
        self.ma_coefs = ArrayHelper.to_array(ma_coefs)

        if self.components.p > 0 and len(self.ar_coefs) == 0:
            self.ar_coefs = np.asarray([0] * self.components.p)

        if self.components.q > 0 and len(self.ma_coefs) == 0:
            self.ma_coefs = np.asarray([0] * self.components.q)

    def __init_x0(self, x0=None):
        self.x0 = x0
        if self.x0 is None:
            self.x0 = self._create_x0_of_zeroes()



import numpy as np
import copy

from ..abstract import ArrayHelper, Components as AbstractComponents


class Components(AbstractComponents):
    """Contains information necessary to determine amount of parameters of the model.

    See parent class description for details.
    """

    def __init__(self, use_box_cox=False, use_trend=False, use_damped_trend=False,
                 seasonal_periods=None, seasonal_harmonics=None,
                 use_arma_errors=False, p=0, q=0,
                 box_cox_bounds=(0, 1)):
        """
        Components class contains all the information necessary to determine the amount and the structure
        of parameters used in the model but not parameters' values.

        Parameters
        ----------
        seasonal_harmonics: array-like of positive int values
            Amount of harmonics used to model each of the seasons.
            Length of this array must equal amount of seasons (length of seasonal_periods)
        """
        super().__init__(
            use_box_cox=use_box_cox, use_trend=use_trend, use_damped_trend=use_damped_trend,
            seasonal_periods=seasonal_periods,
            use_arma_errors=use_arma_errors, p=p, q=q,
            box_cox_bounds=box_cox_bounds)

        self._init_seasonal_harmonics(seasonal_harmonics)
        # assertion: all provided harmonics are positive, 0 should not be provided

    @classmethod
    def create_constant_components(cls):
        return cls(use_box_cox=False, use_trend=False, seasonal_periods=None, use_arma_errors=False)

    def gamma_params_amount(self):
        """TBATS model contains two gamma parameters for each seasonality"""
        return 2 * len(self.seasonal_periods)

    def seasonal_components_amount(self):
        """TBATS model contains two seed values for each seasonality being modelled"""
        return self.seasonal_harmonics * 2

    def with_harmonics_as_ones(self):
        """Creates copy of itself but with all harmonics set to 1

        Returns
        -------
        Components
            copy of components with all harmonics set to 1
        """
        me = copy.deepcopy(self)
        me._init_seasonal_harmonics()  # sets them to a vector of ones
        return me

    def with_harmonic_for_season(self, season_index, new_harmonic):
        """Creates copy of itself with provided harmonics amount for chosen season.

        Parameters
        ----------
        season_index: int
            Index of the season to change harmonic for.
            Index for first season is 0.
        new_harmonic: int
            New amount of harmonics for this season.
            Only positive values are allowed.

        Returns
        -------
        Components
            copy of components with provided harmonics amount for chosen season.
        """
        me = copy.deepcopy(self)
        me.seasonal_harmonics[season_index] = new_harmonic
        return me

    def with_seasonal_periods(self, seasonal_periods):
        """Creates copy of itself but with new seasonal periods

        All harmonics will be set to 1

        Parameters
        ----------
        seasonal_periods: array-like
            New season lengths

        Returns
        -------
        Components
            copy of components with new seasonal periods
        """
        me = super().with_seasonal_periods(seasonal_periods)
        me._init_seasonal_harmonics()
        return me

    def without_seasonal_periods(self):
        """Creates copy itself without seasonality

        Returns
        -------
        Components
            copy of components without seasonal periods
        """
        me = super().without_seasonal_periods()
        me._init_seasonal_harmonics()
        return me

    def _normalize_seasons(self, seasonal_periods):
        """Ensures all seasons are float values"""
        return ArrayHelper.to_array(seasonal_periods, float)

    def _init_seasonal_harmonics(self, seasonal_harmonics=None):
        self.seasonal_harmonics = ArrayHelper.to_array(seasonal_harmonics, int)
        if len(self.seasonal_harmonics) != len(self.seasonal_periods):
            self.seasonal_harmonics = np.asarray([1] * len(self.seasonal_periods), int)

    def _seasonal_summary(self):
        return "Seasonal harmonics %s\n" % self.seasonal_harmonics



import numpy as np
from scipy.optimize import minimize

import tbats.error as error


class ParamsOptimizer(object):
    """ Tries to find optimal seed and parameter values for the model

    Uses linear regression to choose seed values
    and Nelder-Mead optimization to choose model parameters

    Attributes
    ----------
    optimal_params: ModelParams or None
        Optimal model parameters that were found.
        None when no optimization was performed.

    Methods
    -------
    optimize(y, initial_params, calculate_seed=True)
        Seeks for the best seed and parameter values for the model defined by initial params
    converged()
        Returns true when optimization did converge
    optimal_model()
        Returns the best model that was found
    """

    def __init__(self, context):
        self.context = context

        # Initialize optimization params
        self.optimal_params = None  # no parameters found yet
        self._success = False  # did not converge yet
        self._y = None
        self._starting_params = None
        self._scale = None

    def converged(self):
        """Returns true when optimization converged

        Returns
        -------
        bool
            True when converged
        """
        return self._success

    def optimal_model(self):
        """Constructs the best model that was found

        Returns
        -------
        Model
            Optimal model
        """
        if self._y is None:
            self.context.get_exception_handler().exception("No optimization was performed yet", error.BatsException)
        # will return a model even if optimization did not converge
        model = self.context.create_model(self.optimal_params).fit(self._y)
        if not model.is_fitted:
            model.add_warning("Model did not calculate properly and seems unusable.")
        if not model.can_be_admissible():
            model.add_warning("Model is not admissible! Forecasts may be unstable. Check long term forecasts.")
        if not self.converged():
            model.add_warning("Optimization did not converge. Forecasts may be unstable.")
        return model

    def optimize(self, y, initial_params, calculate_seed=True):
        """Finds seed and parameters that minimize likelihood

        Parameters
        ----------
        y: array-like of floats
            Time series
        initial_params: ModelParams
            Model components and initial parameter values
        calculate_seed: bool, optional (default=True)
            If optimizer should also calculate seed x0.
            When False optimization shall use x0 from initial params as seed

        Returns
        -------
        ParamsOptimizer
            Itself
        """
        self._y = y
        self._scale = None  # reset scale

        # Initialize seed using linear regression
        self._starting_params = initial_params
        if calculate_seed:
            x0 = self._calculate_seed_x0(y, initial_params)
            self._starting_params = initial_params.with_x0(x0)

        starting_vector = self._starting_params.to_vector()
        result = minimize(
            self._scale_and_calculate_likelihood,  # function being optimized
            x0=self._inv_scale_vector(starting_vector),
            # self.calculate_likelihood,
            # x0=starting_vector,
            method='Nelder-Mead',
            options={
                'maxiter': 100 * (len(starting_vector) ** 2),
                'fatol': 1e-8,
                # 'disp': True,
            }
        )
        self._success = result.success

        self.optimal_params = self._starting_params.with_vector_values(
            self._scale_vector(result.x)
        )

        return self

    def _scale_and_calculate_likelihood(self, optimization_vector):
        """Scales parameters vector (for performance) and calculates model likelihood"""
        optimization_vector = self._scale_vector(optimization_vector)
        return self._calculate_likelihood(optimization_vector)

    def _calculate_likelihood(self, optimization_vector):
        """Fits model with provided parameters vector to time series and returns its likelihood"""
        infinity = 10 ** 10  # we can not return np.inf as optimization will not work
        # print(optimization_vector)
        params = self._starting_params.with_vector_values(optimization_vector)
        model = self.context.create_model(params, validate_input=False)
        if not model.can_be_admissible():  # don't even calculate the model for such params
            return infinity
        model = model.fit(self._y)
        likelihood = model.likelihood()
        if likelihood == np.inf:
            return infinity
        return likelihood

    def _calculate_seed_x0(self, y, params):
        """Calculates seed values using seed finder (linear regression)"""
        model = self.context.create_model(params.with_zero_x0(), validate_input=False)
        y_tilda = model.fit(y).resid_boxcox

        w = model.matrix.make_w_vector()

        D = model.matrix.calculate_D_matrix()

        w_tilda = np.zeros((len(y), len(w)))
        w_tilda[0, :] = w

        D_transposed = D.T
        for t in range(1, len(y)):
            w_tilda[t, :] = D_transposed @ w_tilda[t - 1, :]

        # TODO I am wondering, since we are removing this ARMA part, do we need to calculate it

        # linear regression to find x0
        seed_finder = self.context.create_seed_finder(params.components)
        return seed_finder.find(w_tilda, y_tilda)

    def _scale_vector(self, vector):
        """Scales vector"""
        scale = self._get_scale()
        return vector * scale

    def _inv_scale_vector(self, vector):
        """Scales vector back"""
        scale = self._get_scale()
        return vector / scale

    def _get_scale(self):
        """Provides vector value scaling factors used to speed-up optimization"""
        if self._scale is not None:
            return self._scale
        components = self._starting_params.components
        scale = [[0.01]]  # alpha scale
        if components.use_box_cox:
            scale.append([0.001])  # box-cox lambda scale
        if components.use_trend:
            scale.append([0.01])  # beta scale
            if components.use_damped_trend:
                scale.append([0.01])  # phi scale
        if components.gamma_params_amount() > 0:
            scale.append([1e-5] * components.gamma_params_amount())  # seasonal parameters scale
        if components.arma_length():
            scale.append([0.1] * components.arma_length())  # ARMA parameters scale
        self._scale = np.concatenate(scale)
        return self._scale



import numpy as np
from sklearn.utils.validation import check_array, column_or_1d as c1d
import scipy.stats as stats
from numpy.linalg import LinAlgError
import warnings

import tbats.error as error
import tbats.transformation as transformation


class Model(object):
    """BATS or TBATS model

    Attributes
    ----------
    warnings: array-like
        All warning messages associated to the model.
        Empty array when there are no warnings.
    is_fitted: bool
        If the model has been successfully fitted
        False when model was not fitted at all or fitting of the model failed
    y: array-like or None
        Time series model has been fitted to.
        None when no fitting was performed yet.
    y_hat: array-like or None
        In series predictions.
    resid: array-like or None
        Residuals from in series predictions
    resid_boxcox: array-like or None
        Residuals from Box-Cox transformed predictions.
        When no box cox is used those are equal to resid.
    x_last: array-like or None
        last vector from x matrix that can be used to calculate predictions
    aic: float
        AIC criterion value
        np.inf when not fitted or failed to fit
    params: ModelParams
        Parameters used in the model
    matrix: MatrixBuilderInterface
        Matrices used in linear equations of the model

    Methods
    -------
    fit(y)
        Calculates in-series predictions for provided time series.
        Can also be used to re-fit the model.
    forecast(steps=1)
        Calculates forecast for provided amount of steps ahead.
    """

    def __init__(self, model_params, context, validate_input=True):
        """Prepares model for fitting.

        Do not use this constructor directly. See abstract.Context methods.

        Parameters
        ----------
        model_params: abstract.ModelParams
            All model parameters and components needed for calculation.
        context: abstract.ContextInterface
            The context used for implementation details.
        validate_input: bool
            If input time series y should be trusted to be valid or not.
            Used for performance.
        """
        self.context = context
        self.warnings = []
        self.params = model_params
        self.validate_input = validate_input
        self.matrix = self.context.create_matrix_builder(self.params)

        self.is_fitted = False
        self.y = None
        self.y_hat = None
        self.resid_boxcox = None
        self.resid = None
        self.x_last = None
        self.aic = np.inf

    def fit(self, y):
        """Calculates in series predictions for provided time series.

        Parameters
        ----------
        y: array-like
            Time series to fit model to

        Returns
        -------
        self
            See class attributes to read fit results
        """
        return self._fit_to_observations(y, self.params.x0)

    def forecast(self, steps=1, confidence_level=None):
        """Forecast for provided amount of steps ahead

        When confidence_level it will also return confidence bounds. Their calculation is valid
        under assumption that residuals are from normal distribution.

        Parameters
        ----------
        steps: int
            Amount of steps to forecast
        confidence_level: float, optional (default=None)
            When provided and a value between 0 and 1 also confidence bounds shall be returned for provided level.
            If None (default), confidence intervals will not be calculated.

        Returns
        -------
        array-like:
            Forecasts
        dict:
            Confidence bounds, present only when confidence_level is provided
        """
        if not self.is_fitted:
            self.context.get_exception_handler().exception(
                'Model must be fitted to be able to forecast. Use fit method first.',
                error.BatsException
            )
        steps = int(steps)
        if steps < 1:
            self.context.get_exception_handler().exception(
                'Parameter \'steps\' must be a positive integer',
                error.InputArgsException
            )

        F = self.matrix.make_F_matrix()
        w = self.matrix.make_w_vector()

        # initialize matrices
        yw_hat = np.asarray([0.0] * steps)
        x = self.x_last

        for t in range(0, steps):
            yw_hat[t] = w @ x
            x = F @ x

        y_hat = self._inv_boxcox(yw_hat)

        if confidence_level is None:
            return y_hat

        return y_hat, self._calculate_confidence_intervals(y_hat, confidence_level)

    def summary(self):
        """Returns model summary containing all parameter values.

        Returns
        -------
        str
            Model summary
        """
        str = ''
        str += self.params.summary() + '\n'
        str += 'AIC %f' % self.aic
        return str

    def likelihood(self):
        """Calculates likelihood of the model. Used for optimization."""
        if not self.is_fitted:
            return np.inf

        residuals = self.resid_boxcox

        with warnings.catch_warnings():
            warnings.filterwarnings('error')
            try:
                likelihood_part = len(residuals) * np.log(np.sum(residuals * residuals))
            except RuntimeWarning:
                # calculation issues, values close to max float value
                return np.inf

        boxcox_part = 0
        if self.params.components.use_box_cox:
            boxcox_part = 2 * (self.params.box_cox_lambda - 1) * np.sum(np.log(self.y))

        return likelihood_part - boxcox_part

    def calculate_aic(self):
        """Calculates AIC criterion value for the model"""
        likelihood = self.likelihood()
        if likelihood == np.inf:
            return np.inf
        return likelihood + 2 * self.params.amount()

    def _fit_to_observations(self, y, starting_x):
        """Fits model with starting x to time series"""
        self.warnings = []
        self.is_fitted = False

        if self.validate_input:
            try:
                y = c1d(check_array(y, ensure_2d=False, force_all_finite=True, ensure_min_samples=1,
                                    copy=True, dtype=np.float64))  # type: np.ndarray
            except Exception as validation_exception:
                self.context.get_exception_handler().exception("y series is invalid",
                                                               error.InputArgsException,
                                                               previous_exception=validation_exception)
        self.y = y
        yw = self._boxcox(y)

        matrix_builder = self.matrix
        w = matrix_builder.make_w_vector()
        g = matrix_builder.make_g_vector()
        F = matrix_builder.make_F_matrix()

        # initialize matrices
        yw_hat = np.asarray([0.0] * len(y))
        # x = np.matrix(np.zeros((len(params.x0), len(yw) + 1)))
        x = starting_x

        with warnings.catch_warnings():
            warnings.filterwarnings('error')
            try:
                for t in range(0, len(y)):
                    yw_hat[t] = w @ x
                    e = yw[t] - yw_hat[t]
                    x = F @ x + g * e
            except RuntimeWarning:
                # calculation issues, values close to max float value
                self.add_warning('Numeric calculation issues detected. Model is not usable.')
                self.is_fitted = False
                return self

        # store fit results
        self.x_last = x
        self.resid_boxcox = yw - yw_hat
        try:
            self.y_hat = self._inv_boxcox(yw_hat)
        except RuntimeWarning:
            self.add_warning('Box-Cox related numeric calculation issues detected. Model is not usable.')
            self.is_fitted = False
            return self

        self.resid = self.y - self.y_hat

        self.is_fitted = True
        self.aic = self.calculate_aic()

        return self

    def _boxcox(self, y):
        yw = y
        if self.params.components.use_box_cox:
            yw = transformation.boxcox(y, lam=self.params.box_cox_lambda)
        return yw

    def _inv_boxcox(self, yw):
        y = yw
        if self.params.components.use_box_cox:
            y = transformation.inv_boxcox(yw, lam=self.params.box_cox_lambda, force_valid=True)
        return y

    def _calculate_confidence_intervals(self, predictions, level):
        """Calculates confidence intervals

        Parameters
        ----------
        predictions: array-like
            Predictions from the model
        level: float
            Confidence level as a number between 0 and 1.
            For example 0.95 denotes 95% confidence interval.

        Returns
        -------
        dict
            mean (predictions), lower_bound, upper_bound, std, calculated_for_level
        -------

        """
        F = self.matrix.make_F_matrix()
        g = self.matrix.make_g_vector()
        w = self.matrix.make_w_vector()

        c = np.asarray([1.0] * len(predictions))
        f_running = np.identity(F.shape[1])
        for step in range(1, len(predictions)):
            c[step] = w @ f_running @ g
            f_running = f_running @ F
        variance_multiplier = np.cumsum(c * c)

        base_variance_boxcox = np.sum(self.resid_boxcox * self.resid_boxcox) / len(self.y)
        variance_boxcox = base_variance_boxcox * variance_multiplier
        std_boxcox = np.sqrt(variance_boxcox)

        z_scores = std_boxcox * np.abs(stats.norm.ppf((1 - level) / 2))
        lower_bound_boxcox = self._boxcox(predictions) - z_scores
        upper_bound_boxcox = self._boxcox(predictions) + z_scores

        return dict(
            mean=predictions,
            lower_bound=self._inv_boxcox(lower_bound_boxcox),
            upper_bound=self._inv_boxcox(upper_bound_boxcox),
            calculated_for_level=level,
        )

    def can_be_admissible(self):
        """Tells if model can be admissible."""
        if not self.params.is_box_cox_in_bounds():
            return False

        params = self.params
        if params.components.use_damped_trend and (params.phi < 0.8 or params.phi > 1):
            return False

        if not self.__AR_is_stationary(params.ar_coefs):
            return False

        if not self.__MA_is_invertible(params.ma_coefs):
            return False

        D = self.matrix.calculate_D_matrix()
        return self.__D_matrix_eigen_values_check(D)

    def is_admissible(self):
        """Tells if model is admissible (stable). Model that has not been fitted is not addmisible."""
        if not self.is_fitted:
            return False
        return self.can_be_admissible()

    @staticmethod
    def __AR_is_stationary(ar_coefs):
        # cut out trailing and non-significant AR components
        significant_indices = np.where(np.abs(ar_coefs) > 1e-08)[0]
        if len(significant_indices) == 0:
            return True
        p = np.max(significant_indices) + 1
        significant_ar_coefs = ar_coefs[0:p]

        roots = np.polynomial.polynomial.polyroots(np.concatenate([[1], -significant_ar_coefs]))
        # Note that np.abs also works with complex numbers. It provides length of complex number
        # There should be no roots in the unit circle
        return np.all(np.abs(roots) > 1.0)

    @staticmethod
    def __MA_is_invertible(ma_coefs):
        # cut out trailing and non-significant AR components
        significant_indices = np.where(np.abs(ma_coefs) > 1e-08)[0]
        if len(significant_indices) == 0:
            return True
        q = np.max(significant_indices) + 1
        significant_ma_coefs = ma_coefs[0:q]

        roots = np.polynomial.polynomial.polyroots(np.concatenate([[1], significant_ma_coefs]))
        # Note that np.abs also works with complex numbers. It provides length of complex number
        # There should be no roots in the unit circle
        return np.all(np.abs(roots) > 1.0)

    @staticmethod
    def __D_matrix_eigen_values_check(D):
        try:
            eigen_values = np.linalg.eigvals(D)
        except LinAlgError:
            return False
        return np.all(np.abs(eigen_values) < 1.01)

    def add_warning(self, message):
        """Add a warning message to the model

        Parameters
        ----------
        message: str
            The message
        """
        self.warnings.append(message)



import numpy as np

from ..abstract import ArrayHelper, ModelParams as AbstractModelParams


class ModelParams(AbstractModelParams):
    """Holds all parameters needed to calculate TBATS model and forecasts

    For other parameters documentation see parent class documentation.

    Parameters
    ----------
    gamma_params: array-like of floats
        Gamma smoothing parameters for seasonal effects.
        In TBATS there are two smoothing parameters for each season.
        Vector holds 2 gamma parameters for season 1, then 2 gamma parameters for season 2, ...
    """

    def __init__(self, components, alpha, beta=None, phi=None, box_cox_lambda=None,
                 gamma_params=None,
                 ar_coefs=None, ma_coefs=None, x0=None):
        """See parent class documentation for details"""
        super().__init__(components=components, alpha=alpha, beta=beta, phi=phi, box_cox_lambda=box_cox_lambda,
                         gamma_params=gamma_params, ar_coefs=ar_coefs, ma_coefs=ma_coefs, x0=x0)

    @classmethod
    def with_default_starting_params(cls, y, components):
        """Factory method for starting model parameters.

        Parameters
        ----------
        y: array-like of floats
            Time series
        components: Components
            Components used in the model

        Returns
        -------
        ModelParams
            Default starting params.
        """
        alpha = 0.09

        beta = None
        phi = None
        if components.use_trend:
            beta = 0.05
            phi = 1
            if components.use_damped_trend:
                phi = 0.999

        # Here TBATS differs from BATS as staring gamma parameters are zeroes for TBATS
        gamma_params = None
        if len(components.seasonal_periods) > 0:
            gamma_params = [0.0] * components.gamma_params_amount()

        box_cox_lambda = cls.find_initial_box_cox_lambda(y, components)

        # note that ARMA and x0 will be initialized to 0's in constructor
        return cls(components=components,
                   alpha=alpha, beta=beta, phi=phi,
                   gamma_params=gamma_params,
                   box_cox_lambda=box_cox_lambda)

    def seasonal_components_amount(self):
        """TBATS requires this many seed state values for seasonalities"""
        return int(2 * self.components.seasonal_harmonics.sum())

    def gamma_1(self):
        """Gamma 1 smoothing parameters are kept in even position of gamma vector"""
        return self.gamma_params[range(0, len(self.gamma_params), 2)]

    def gamma_2(self):
        """Gamma 2 smoothing parameters are kept in odd position of gamma vector"""
        return self.gamma_params[range(1, len(self.gamma_params), 2)]

    def _normalize_gamma_params(self, gamma_params=None):
        """Ensures gamma_params is an array. If necessary it is initialized to zeroes."""
        gamma_params = ArrayHelper.to_array(gamma_params, float)
        expected_params_amount = 2 * len(self.components.seasonal_periods)
        if len(gamma_params) != expected_params_amount:
            gamma_params = np.asarray([0.0] * expected_params_amount, float)
        return gamma_params



import numpy as np

from ..abstract import ArrayHelper, MatrixBuilder as AbstractMatrixBuilder


class MatrixBuilder(AbstractMatrixBuilder):

    def __init(self, model_params):
        super().__init__(model_params)

    def make_seasonal_components_for_w(self):
        seasonal_components = []
        for period in self.params.components.seasonal_periods:
            seasonal_components.append(ArrayHelper.make_one_and_zeroes_vector(period, one_position='end'))
        if len(seasonal_components) > 0:
            seasonal_components = np.concatenate(seasonal_components)
        return seasonal_components

    def make_gamma_vector(self):
        gamma_params = self.params.gamma_params
        seasonal_periods = self.params.components.seasonal_periods
        gamma_vectors = []
        i = 0
        for gamma in gamma_params:
            gamma_vectors.append(ArrayHelper.make_one_and_zeroes_vector(seasonal_periods[i], one=gamma))
            i += 1
        if len(gamma_vectors) > 0:
            gamma_vectors = np.concatenate(gamma_vectors)
        return np.asarray(gamma_vectors)

    def make_A_matrix(self):
        seasonal_periods = self.params.components.seasonal_periods
        if len(seasonal_periods) == 0:
            return np.zeros((0, 0))
        A_array = []
        for period in seasonal_periods:
            Ai = np.block([
                [np.zeros((1, period - 1)), 1],
                [np.identity(period - 1), np.zeros((period - 1, 1))]
            ])
            A_array.append(Ai)
        tao = int(np.sum(seasonal_periods))
        dsum = np.zeros((tao, tao))
        offsetRow = 0
        offsetCol = 0
        for Ai in A_array:
            dsum[offsetRow:offsetRow + Ai.shape[0], offsetCol:offsetCol + Ai.shape[1]] = Ai
            offsetRow += Ai.shape[0]
            offsetCol += Ai.shape[1]
        return dsum



import numpy as np
from scipy.optimize import minimize_scalar


# Guerrero, V.M. (1993) Time-series analysis supported by power
# transformations. \emph{Journal of Forecasting}, \bold{12}, 37--48.
class Guerrero(object):

    def __init__(self):
        self.y = None
        self.season_length = None

    def find_lambda(self, y, seasonal_periods=None, bounds=(-1, 2)):
        # assertion seasonal_periods are sorted and at least 2
        if seasonal_periods is None:
            seasonal_periods = []
        seasonal_periods = np.asarray(seasonal_periods)

        self.season_length = self.find_longest_season_with_2x_observations(y, seasonal_periods)
        if len(y) < 2 * self.season_length:
            return 1.0

        if np.any(y <= 0):
            bounds[0] = np.max([bounds[0], 0])

        if bounds[0] >= bounds[1]:
            return bounds[0]

        return self.minimize(y, bounds)

    @staticmethod
    def find_longest_season_with_2x_observations(y, seasonal_periods):
        season_length = 2
        for length in seasonal_periods:
            if len(y) < int(length) * 2:
                break
            # there are at least 2 seasons of observations
            season_length = length
        return int(season_length)

    def minimize(self, y, bounds):
        self.y = y
        result = minimize_scalar(
            self.guerrero_coefficient_of_variation,
            method='bounded',
            bounds=bounds,
            options=dict(
                xatol=1e-8,
            )

        )
        return result.x

    def guerrero_coefficient_of_variation(self, lam):
        # assertion y is long enough to have 2 full seasons of observations
        full_seasons = int(len(self.y) / self.season_length)
        non_full_season_offset = int(len(self.y) - full_seasons * self.season_length)
        # nth seasonal moment - nth observation in each season
        y_seasonal_moments = np.reshape(self.y[non_full_season_offset:], (full_seasons, self.season_length))
        y_moment_means = np.mean(y_seasonal_moments, axis=1)
        y_moment_stds = np.std(y_seasonal_moments, axis=1, ddof=1)
        y_moment_ratios = y_moment_stds / y_moment_means ** (1 - lam)
        return np.std(y_moment_ratios, ddof=1) / np.mean(y_moment_ratios)



import numpy as np
from sklearn.linear_model import LinearRegression

from . import Components


class SeedFinder(object):

    def __init__(self, components):
        """
        :param Components components:
        :return:
        """
        self.components = components

    def to_matrix_for_linear_regression(self, w_tilda):
        # abstract method
        raise NotImplementedError()

    def from_linear_regression_coefs_to_x0(self, linear_regression_coefs):
        # abstract method
        raise NotImplementedError()

    def find(self, w_tilda, residuals):
        w_for_lr = self.to_matrix_for_linear_regression(w_tilda)

        # this makes sure that coefficient for all zeroes dimension will be zero
        # without this calculated coefficient may be very large and unrealistic
        for i in range(w_for_lr.shape[1]):
            if np.allclose(w_for_lr[:, i], 0):
                w_for_lr[:, i] = [0] * len(w_for_lr)

        linear_regression = LinearRegression(fit_intercept=False)
        coefs = np.asarray(linear_regression.fit(w_for_lr, residuals).coef_)
        return self.from_linear_regression_coefs_to_x0(coefs)



import numpy as np


class ArrayHelper(object):
    @classmethod
    def to_array(cls, values=None, type=float, give_shape=False):
        if values is None:
            if give_shape:
                return np.zeros((0, 0))
            return np.asarray([])
        if isinstance(values, int) or isinstance(values, float):
            return np.asarray([values], type)
        return np.asarray(values, type)

    @classmethod
    def make_one_and_zeroes_vector(cls, size, one=1, one_position='begin'):
        # makes vector of the form [one,0,0,0]
        v = np.asarray([0.0] * size, dtype=float)

        if one_position == 'begin':
            one_position = 0
        else:  # 'end'
            one_position = size - 1

        if size > 0:
            v[one_position] = one
        return v



from pmdarima.arima import auto_arima


class Case(object):
    """Provides means to choose best model for provided components setup

    Attributes
    ----------
    components: Components
        Components model should use
    model: Model or None
        Best model found.
        None when no fitting occured yet.

    Methods
    """

    def __init__(self, components, context):
        """Creates new BATS/TBATS case

        Don't use this constructor directly. See ContextInterface.create_case

        Parameters
        ----------
        components: Components
            Components the case applies to
        context: ContextInterface
            Used to build models
        """
        self.components = components
        self.context = context
        self.model = None  # not fitted yet

    def fit(self, y):
        """ Fits and chooses best model with chosen components

        Parameters
        ----------
        y: array-like
            Time series to fit the model to and rate the model against

        Returns
        -------
        Model
            Best model by AIC
        """
        best_model = self.fit_initial_model(y)

        if self.components.use_arma_errors:
            # Try adding ARMA to the model
            arma_model = auto_arima(best_model.resid, stationary=True, trend='n',
                                    suppress_warnings=True, error_action='ignore')
            p = arma_model.order[0]
            q = arma_model.order[2]
            if p > 0 or q > 0:  # Found ARMA components
                # Fit model with ARMA errors modelling
                model_candidate = self.fit_case(y, best_model.params.components.with_arma(p, q))
                if model_candidate.aic < best_model.aic:
                    best_model = model_candidate

        self.model = best_model
        return self.model

    def fit_initial_model(self, y):
        # abstract method
        raise NotImplementedError()

    def fit_case(self, y, case_definition):
        """Fits model with provided components to time series

        Parameters
        ----------
        y: array-like
            Time series
        case_definition: Components
            Components for which model should be fit

        Returns
        -------
        Model
            Fitted model
        """
        starting_params = self.context.create_default_starting_params(y, components=case_definition)
        return self.fit_with_starting_params(y, starting_params)

    def fit_with_starting_params(self, y, model_params):
        """ Fits (optimizes) model with provided starting parameters

        Parameters
        ----------
        y: array-like
            Time series
        model_params: ModelParams
            Model parameters

        Returns
        -------
        Model
            optimal model
        """
        optimization = self.context.create_params_optimizer()
        return optimization.optimize(y, model_params).optimal_model()



from ..abstract import MatrixBuilderCache, Context as AbstractContext
from tbats.tbats import *


class Context(AbstractContext):

    def create_constant_model(self, constant_value):
        return self.create_model(
            ModelParams(components=Components.create_constant_components(), alpha=0, x0=[constant_value])
        )

    def create_model(self, params, validate_input=True):
        return Model(
            params,
            validate_input=validate_input,
            context=self,
        )

    def create_seed_finder(self, components):
        return SeedFinder(components)

    def create_matrix_builder(self, params):
        base_builder = MatrixBuilder(params)
        return MatrixBuilderCache(base_builder)

    def create_default_starting_params(self, y, components):
        return ModelParams.with_default_starting_params(y, components=components)

    def create_params_optimizer(self):
        return ParamsOptimizer(self)

    def create_components(self, **components):
        return Components(**components)

    def create_case_from_dictionary(self, **components_dictionary):
        return self.create_case(components=self.create_components(**components_dictionary))

    def create_case(self, components):
        return Case(components=components, context=self)

    def create_harmonics_choosing_strategy(self):
        return HarmonicsChoosingStrategy(self, checking_range=self.n_jobs)


from ..abstract import ArrayHelper, Components as AbstractComponents


class Components(AbstractComponents):
    """Contains information necessary to determine amount of parameters of the model.

    See parent class description for details.
    """
    def __init__(self, use_box_cox=False, use_trend=False, use_damped_trend=False,
                 seasonal_periods=None, use_arma_errors=False, p=0, q=0,
                 box_cox_bounds=(0, 1)):
        super().__init__(
            use_box_cox=use_box_cox, use_trend=use_trend, use_damped_trend=use_damped_trend,
            seasonal_periods=seasonal_periods,
            use_arma_errors=use_arma_errors, p=p, q=q,
            box_cox_bounds=box_cox_bounds)

    @classmethod
    def create_constant_components(cls):
        return cls(use_box_cox=False, use_trend=False, seasonal_periods=None, use_arma_errors=False)

    def seasonal_components_amount(self):
        """Returns amount of seasonal seed states per season"""
        return self.seasonal_periods

    def gamma_params_amount(self):
        """BATS model contains one gamma parameter for each season"""
        return len(self.seasonal_periods)

    def _normalize_seasons(self, seasonal_periods):
        """Ensures seasons are integer values"""
        return ArrayHelper.to_array(seasonal_periods, int)

    def _seasonal_summary(self):
        return ""



import warnings

from . import BatsWarning


class ExceptionHandler(object):

    def __init__(self, show_warnings=True):
        self.show_warnings = show_warnings

    def warn(self, message, warning_type=BatsWarning):
        if self.show_warnings:
            warnings.warn(message, warning_type)
            return True
        return False

    def exception(self, message, exception_type, previous_exception=None):
        if previous_exception is not None:
            raise exception_type(message) from previous_exception
        raise exception_type(message)



from . import ContextInterface
import tbats.error as error
import multiprocessing as actual_processing
import multiprocessing.dummy as dummy_processing

class Context(ContextInterface):

    def __init__(self, show_warnings=True, n_jobs=None, multiprocessing_start_method=None):
        self.exception_handler = error.ExceptionHandler(show_warnings)
        self.n_jobs = n_jobs
        self.multiprocessing_start_method = multiprocessing_start_method

    def get_exception_handler(self):
        return self.exception_handler

    def multiprocessing(self):
        if self.n_jobs == 1:
            return dummy_processing
        if self.multiprocessing_start_method is None:
            return actual_processing
        return actual_processing.get_context(self.multiprocessing_start_method)


import numpy as np


class ComponentMatrix(object):

    def __init__(self, matrix, model_components, append_arma_coefs=False):
        self.components = model_components
        self.matrix = np.atleast_2d(matrix)
        if append_arma_coefs and self.components.arma_length() > 0:
            self.matrix = np.append(self.matrix, np.zeros((len(self.matrix), self.components.arma_length())), axis=1)

    def without_arma(self):
        return __class__(self.matrix[:, :(self.arma_offset())], self.components.without_arma())

    def with_replaced_seasonal(self, new_seasonal_part, new_seasonal_periods):
        if len(new_seasonal_part) == 0 or new_seasonal_part.shape[0] == 0:
            return __class__(
                np.append(self.alpha_beta_part(), self.arma_part(), axis=1),
                self.components.with_seasonal_periods(new_seasonal_periods)
            )
        new_seasonal_part = np.atleast_2d(new_seasonal_part)
        return __class__(
            np.block([self.alpha_beta_part(), new_seasonal_part, self.arma_part()]),
            self.components.with_seasonal_periods(new_seasonal_periods)
        )

    def as_matrix(self):
        return self.matrix

    def as_vector(self):
        """
        :return: First row as flat vector
        """
        return np.asarray(self.matrix.T[:, 0])

    def seasonal_part(self):
        return self.matrix[:, self.seasonal_offset():self.arma_offset()]

    def alpha_beta_part(self):
        return self.matrix[:, :self.seasonal_offset()]

    def arma_part(self):
        return self.matrix[:, self.arma_offset():]

    def break_into_seasons(self):
        w_seasons = []
        offset = 0
        matrix_seasonal_part = self.seasonal_part()
        season_lengths = self.components.seasonal_components_amount()
        for s in season_lengths:
            w_seasons.append(matrix_seasonal_part[:, offset:(offset + s)])
            offset += s
        return w_seasons

    def arma_offset(self):
        return self.matrix.shape[1] - self.components.arma_length()

    def seasonal_offset(self):
        offset = 1
        if self.components.use_trend:
            offset += 1
        return offset
